@@ -34,7 +34,14 @@
 		    (when (and (>= sym-len text-len)
 			       (string=? text (substring sym 0 text-len)))
 		      (if match
-			  (return text)
+			  ;; more than one match, save the longest text that all syms match
+			  (let ((min-len (min (string-length match) sym-len)))
+			    (do ((i text-len (+ i 1)))
+				((or (= i min-len)
+				     (not (char=? (match i) (sym i))))
+				 (if (= min-len text-len)
+				     (return text)
+				     (set! match (substring match 0 i))))))
 			  (set! match sym)))))
 		st)
 	       (or match text)))))
@@ -238,22 +245,22 @@
 			 (set! (hook-functions *unbound-variable-hook*) old-unbound-var-hook))))
 		
 		  (let ((new-load 
-			 (let ((documentation "this is the repl's load replacement; its default is to use the repl's top-level-let.")
-			       (signature '(values string? let?)))
+			 (let ((+documentation+ "this is the repl's load replacement; its default is to use the repl's top-level-let.")
+			       (+signature+ '(values string? let?)))
 			   (lambda* (file (e (*repl* 'top-level-let)))
-			     (dynamic-wind original-hooks (lambda () (load file e)) repl-hooks))))
+			     (dynamic-wind repl-hooks (lambda () (load file e)) original-hooks))))
 			
 			(new-eval 
-			 (let ((documentation "this is the repl's eval replacement; its default is to use the repl's top-level-let.")
-			       (signature '(values list? let?)))
+			 (let ((+documentation+ "this is the repl's eval replacement; its default is to use the repl's top-level-let.")
+			       (+signature+ '(values list? let?)))
 			   (lambda* (form (e (*repl* 'top-level-let)))
-			     (dynamic-wind original-hooks (lambda () (eval form e)) repl-hooks))))
+			     (dynamic-wind repl-hooks (lambda () (eval form e)) original-hooks))))
 			
 			(new-eval-string 
-			 (let ((documentation "this is the repl's eval-string replacement; its default is to use the repl's top-level-let.")
-			       (signature '(values string? let?)))
+			 (let ((+documentation+ "this is the repl's eval-string replacement; its default is to use the repl's top-level-let.")
+			       (+signature+ '(values string? let?)))
 			   (lambda* (str (e (*repl* 'top-level-let)))
-			     (dynamic-wind original-hooks (lambda () (eval-string str e)) repl-hooks)))))
+			     (dynamic-wind repl-hooks (lambda () (eval-string str e)) original-hooks)))))
 		    
 		    (dynamic-wind
 			(lambda ()
@@ -415,8 +422,10 @@
 	    ;;    disable: (format *stderr* "~C[?9l" #\escape)
 	    ;; while enabled, mouse selection instead sends coords to repl (so it's annoying)
 	    ;;    also it's sticky!  exit repl does not clear this flag so mouse is effectively dead
-	    ;; upon click, we get ESC [ M bxy -- need to poll for this?
-	    
+	    ;; upon click, we get ESC [ M b x y (b = button info, x/y are coords starting at 32=0 -- 6 chars received)
+	    ;; to trap C-C (define (sigcatch no) (tty-reset) (#_exit)) (if (equal? (signal SIGINT sigcatch) SIG_ERR) (#_exit))
+	    ;; to see click, add a meta-key for M
+	    ;; all this apparently only works in xterm, so I hesitate to build on it here.
 	    
 	    ;; -------- display --------
 	    (define (display-prompt)
@@ -507,7 +516,7 @@
 		      ((null? lst))
 		    (let ((str ((car lst) c)))
 		      (move-cursor i col)
-		      (format *stderr* "~C[K| ~A"  #\escape (if (> (length str) col) (substring str 0 (- col 1)) str))))
+		      (format *stderr* "~C[K| ~A" #\escape (if (> (length str) col) (substring str 0 (- col 1)) str))))
 		  
 		  (move-cursor (+ 2 (length (*repl* 'helpers))) col)
 		  (format *stderr* "+~NC" (- col 2) #\-)
@@ -548,29 +557,29 @@
 	    (define keymap (dilambda
 			    (lambda (c)
 			      (cond ((char? c) 
-				     (keymap-functions (char->integer c)))
+				     (vector-ref keymap-functions (char->integer c)))
 				    ((integer? c) 
-				     (keymap-functions c))
+				     (vector-ref keymap-functions c))
 				    ((not (string? c))
 				     (error 'wrong-type-arg "keymap takes a character or string argument"))
 				    ((= (length c) 1) 
-				     (keymap-functions (char->integer (c 0))))
+				     (vector-ref keymap-functions (char->integer (c 0))))
 				    ((and (= (length c) 2)
 					  (char=? (c 0) #\escape))
-				     (meta-keymap-functions (char->integer (c 1))))
+				     (vector-ref meta-keymap-functions (char->integer (c 1))))
 				    (else (lambda (c) #t))))
 			    (lambda (c f)
 			      (cond ((char? c) 
-				     (set! (keymap-functions (char->integer c)) f))
+				     (vector-set! keymap-functions (char->integer c) f))
 				    ((integer? c) 
-				     (set! (keymap-functions c) f))
+				     (vector-set! keymap-functions c f))
 				    ((not (string? c))
 				     (error 'wrong-type-arg "set! keymap takes a character or string first argument"))
 				    ((= (length c) 1)
-				     (set! (keymap-functions (char->integer (c 0))) f))
+				     (vector-set! keymap-functions (char->integer (c 0)) f))
 				    ((and (= (length c) 2)
 					  (char=? (c 0) #\escape))
-				     (set! (meta-keymap-functions (char->integer (c 1))) f))))))
+				     (vector-set! meta-keymap-functions (char->integer (c 1)) f))))))
 	    
 	    (define C-a 1)     ; #\x01 etc
 	    (define C-b 2)
@@ -643,15 +652,15 @@
 	      
 	      (do ((i 0 (+ i 1)))
 		  ((= i 32))
-		(set! (keymap-functions i) no-op-keyfunc))
+		(vector-set! keymap-functions i no-op-keyfunc))
 	      
 	      (do ((i 32 (+ i 1)))
 		  ((= i 256))
-		(set! (keymap-functions i) main-keyfunc))
+		(vector-set! keymap-functions i main-keyfunc))
 	      
 	      (do ((i 0 (+ i 1)))
 		  ((= i 256))
-		(set! (meta-keymap-functions i) no-op-keyfunc)))
+		(vector-set! meta-keymap-functions i no-op-keyfunc)))
 	    
 	    
 	    ;; -------- cursor movement 
@@ -1037,50 +1046,62 @@
 	      (set! (meta-keymap-functions (char->integer #\u)) upper-case)
 	      (set! (meta-keymap-functions (char->integer #\l)) lower-case))
 	    
-	    ;; -------- terminal setup --------
-	    (define* (run file)
-	      (let ((saved #f)
-		    (tty #t))
+	    ;; -------- emacs --------
+	    (define (emacs-repl)
+	      ;; TODO: use the emacs language server protocol? (not our own rpc stuff or epc), for json, see json.scm 
+	      ;;   probably will need an argument/function? for repl to open the channel or whatever
+	      ;; also this does not resend the entire expression after editing
+	      ;;   and does not notice in-place edits
+	      ;;   can <cr> get entire expr?
+	      (let ((buf (c-pointer->string (calloc 512 1) 512)))
+		(format *stderr* "> ")
+		(do ((b (fgets buf 512 stdin) (fgets buf 512 stdin)))
+		    ((zero? (length b))
+		     (#_exit))
+		  (let ((len (strlen buf)))
+		    (when (positive? len)
+		      (do ((i 0 (+ i 1)))
+			  ((or (not (char-whitespace? (buf i)))
+			       (= i len))
+			   (when (< i len)
+			     (let ((str (substring buf 0 (- (strlen buf) 1))))
+			       ;(format *stderr* "str: ~S~%" str)
+			       (catch #t
+				 (lambda ()
+				   (do ()
+				       ((= (string-length str) 0))
+				     (catch 'string-read-error
+				       (lambda ()
+					 (with-repl-let
+					  (lambda ()
+					    (format *stderr* "~S~%> " (eval-string str (*repl* 'top-level-let)))
+					    (set! str ""))))
+				       (lambda (type info)
+					 (fgets buf 512 stdin)
+					 ;(format *stderr* "add str: ~S~%" (substring buf 0 (- (strlen buf) 1)))
+					 (set! str (string-append str " " (substring buf 0 (- (strlen buf) 1))))))))
+				 (lambda (type info)
+				   (set! str "")
+				   (apply format *stderr* info)
+				   (format *stderr* "~%> "))))))))))))
+
+	    ;; -------- rxvt et al --------
+	    (define (terminal-repl file)		
+	      (let ((saved #f))
 		
 		;; we're in libc here, so exit is libc's exit!
 		(define (tty-reset)
-		  (if tty (tcsetattr terminal-fd TCSAFLUSH saved))
+		  (tcsetattr terminal-fd TCSAFLUSH saved)
 		  (if (not (equal? input-fd terminal-fd)) (close input-fd))
 		  (#_exit))
 		
 		(varlet (*repl* 'top-level-let) 
-		  :exit (let ((documentation "(exit) resets the repl tty and exits the repl"))
+		  :exit (let ((+documentation+ "(exit) resets the repl tty and exits the repl"))
 			  (lambda ()
 			    (newline *stderr*)
 			    (tty-reset))))
 		
-		;; check for dumb terminal
-		(if (or (zero? (isatty terminal-fd))        ; not a terminal -- input from pipe probably
-			(string=? (getenv "TERM") "dumb"))  ; no vt100 codes -- emacs shell for example
-		    (let ((buf (c-pointer->string (calloc 512 1) 512)))
-		      (set! tty #f)
-		      (format *stderr* "> ")
-		      (do ((b (fgets buf 512 stdin) (fgets buf 512 stdin)))
-			  ((zero? (length b))
-			   (#_exit))
-			(let ((len (strlen buf)))
-			  (when (positive? len)
-			    (do ((i 0 (+ i 1)))
-				((or (not (char-whitespace? (buf i)))
-				     (= i len))
-				 (when (< i len)
-				   (with-repl-let
-				    (lambda ()
-				      (catch #t
-					(lambda ()
-					  (format *stderr* "~S~%" (eval-string (substring buf 0 (- (strlen buf) 1)) (*repl* 'top-level-let))))
-					(lambda (type info)
-					  (format *stderr* "error: ")
-					  (apply format *stderr* info)
-					  (newline *stderr*)))))
-				   (format *stderr* "> ")))))))))
-		
-		;; not a pipe or a dumb terminal -- hopefully all others accept vt100 codes
+		;; a "normal" terminal -- hopefully it accepts vt100 codes
 		(let ((buf (termios.make)))
 		  (let ((read-size 128))
 		    (set! next-char                                     ; this indirection is needed if user pastes the selection into the repl
@@ -1098,10 +1119,10 @@
 				   (set! chars (read input-fd cc read-size))
 				   (if (= chars 0)
 				       (tty-reset))
-
+				   
 				   (when (> chars (- last-col prompt-length 12))
 				     (let ((str (substring c 0 chars)))
-
+				       
 				       (when (= chars read-size)
 					 ;; concatenate buffers until we get the entire selection
 					 (let reading ((num (read input-fd cc read-size)))
@@ -1157,17 +1178,17 @@
 						(display-lines)
 						(return #\newline)))))
 				       ;; now the pasted-in line has inserted newlines, we hope
-
+				       
 				       (set! c str)
 				       (set! cc (string->c-pointer c))
-
+				       
 				       ;; avoid time-consuming redisplays.  We need to use a recursive call on next-char here
 				       ;;   since we might have multi-char commands (embedded #\escape -> meta, etc)
 				       ;; actually, the time is not the repl's fault -- xterm seems to be waiting
 				       ;;   for the window manager or someone to poke it -- if I move the mouse,
 				       ;;   I get immediate output.  I also get immediate output in any case in OSX.
 				       ;;   valgrind and ps say we're not computing, we're just sitting there.
-
+				       
 				       (catch #t
 					 (lambda ()
 					   (do ((ch (next-char) (next-char)))
@@ -1207,8 +1228,6 @@
 		  (when (negative? (tcsetattr terminal-fd TCSAFLUSH buf))
 		    (tty-reset))
 		  
-		  
-		  ;; -------- the repl --------
 		  (display-prompt)
 		  (cursor-bounds)
 		  ;; (debug-help)
@@ -1237,6 +1256,13 @@
 			(set! chars 0)
 			(new-prompt)))))))
 	    
+	    (define* (run file)
+	      ;; check for dumb terminal
+	      (if (or (zero? (isatty terminal-fd))        ; not a terminal -- input from pipe probably
+		      (string=? (getenv "TERM") "dumb"))  ; no vt100 codes -- emacs subjob for example
+		  (emacs-repl)                            ; TODO: restore support for the pipe case
+		  (terminal-repl file)))
+
 	    (curlet))))))
       
       (define (save-repl) 
@@ -1258,7 +1284,6 @@
 	(set! (*repl* 'top-level-let) (load "save.repl")))  
       ;; I think this could be a merge rather than a reset by using (with-let top-level-let (load ...))
       
-      
       (set! keymap (repl-let 'keymap))
       (set! history (repl-let 'history))
       (set! history-size (repl-let 'history-size))
@@ -1411,7 +1436,7 @@
 					    "~C[38;5;208m~A~C[0m: ~S~%"))                              ; orange for less likely choices
 				    #\escape (caar b) #\escape
 				    (if (procedure? (cdar b))
-					(let ((doc (procedure-documentation (cdar b)))) ; returns "" if no doc
+					(let ((doc (documentation (cdar b)))) ; returns "" if no doc
 					  (if (positive? (length doc))
 					      doc
 					      'procedure))
@@ -1478,7 +1503,7 @@
 
 ;;; to display a variable's value as s7 runs using the repl help window:
 ;;;    (define xyz 1) ; some variable...
-;;;    (set! (symbol-access 'xyz) (lambda (sym val) (set! (*repl* 'helpers) (list (lambda (c) (format #f "xyz: ~S" val)))) val))
+;;;    (set! (symbol-setter 'xyz) (lambda (sym val) (set! (*repl* 'helpers) (list (lambda (c) (format #f "xyz: ~S" val)))) val))
 
 
 ;;; --------------------------------------------------------------------------------
